use std::{ops::Deref, sync::Arc};

use ash::{prelude::VkResult, vk};
use vk_mem::Alloc;

use crate::Device;

pub struct Buffer {
    device: Device,
    buffer: vk::Buffer,
    allocation: vk_mem::Allocation,
    usage: vk::BufferUsageFlags,
    size: u64,
}

impl Drop for Buffer {
    fn drop(&mut self) {
        unsafe {
            self.device
                .alloc()
                .destroy_buffer(self.buffer, &mut self.allocation);
        }
    }
}

impl Buffer {
    pub fn new(
        device: Device,
        size: usize,
        usage: vk::BufferUsageFlags,
        queue_families: &[u32],
        memory_usage: vk_mem::MemoryUsage,
        alloc_flags: vk_mem::AllocationCreateFlags,
    ) -> VkResult<Arc<Self>> {
        let sharing_mode = if queue_families.len() > 1 {
            vk::SharingMode::EXCLUSIVE
        } else {
            vk::SharingMode::CONCURRENT
        };

        let (buffer, allocation) = unsafe {
            device.alloc().create_buffer(
                &vk::BufferCreateInfo::default()
                    .usage(usage)
                    .queue_family_indices(queue_families)
                    .sharing_mode(sharing_mode),
                &vk_mem::AllocationCreateInfo {
                    flags: alloc_flags,
                    usage: memory_usage,
                    ..Default::default()
                },
            )?
        };

        let buffer = Self {
            device,
            buffer,
            allocation,
            usage,
            size: size as u64,
        };

        Ok(Arc::new(buffer))
    }

    pub fn map(&mut self) -> VkResult<MappedBuffer<'_>> {
        let bytes = unsafe {
            let data = self.device.alloc().map_memory(&mut self.allocation)?;
            let slice = core::slice::from_raw_parts(data, self.size as usize);

            slice
        };

        Ok(MappedBuffer { bytes })
    }
}

pub struct MappedBuffer<'a> {
    bytes: &'a [u8],
}

impl Deref for MappedBuffer<'_> {
    type Target = [u8];

    fn deref(&self) -> &Self::Target {
        self.bytes
    }
}
