use std::{future::Future, marker::PhantomData, sync::Arc};

use crate::{
    buffers::Buffer,
    images::{Image2D, QueueOwnership},
    sync::{self, FenceFuture},
    util::{FormatExt, MutexExt},
};

use super::{Device, Queue};
use ash::{prelude::*, vk};
use parking_lot::Mutex;

pub struct SingleUseCommandPool {
    device: Device,
    pool: Mutex<vk::CommandPool>,
    queue: Queue,
}

impl Drop for SingleUseCommandPool {
    fn drop(&mut self) {
        unsafe {
            self.pool.with_locked(|&pool| {
                self.device.dev().destroy_command_pool(pool, None);
            });
        }
    }
}

impl SingleUseCommandPool {
    pub fn new(device: Device, queue: Queue) -> VkResult<Arc<Self>> {
        let pool_info = vk::CommandPoolCreateInfo::default()
            .queue_family_index(queue.family())
            .flags(vk::CommandPoolCreateFlags::TRANSIENT);

        let pool =
            unsafe { device.dev().create_command_pool(&pool_info, None)? };

        Ok(Arc::new(Self {
            device,
            pool: Mutex::new(pool),
            queue,
        }))
    }

    pub fn alloc(self: &Arc<Self>) -> VkResult<SingleUseCommand> {
        SingleUseCommand::new(self.device.clone(), self.clone())
    }

    pub fn queue(&self) -> &Queue {
        &self.queue
    }

    /// get the underlying pool, bypassing the mutex
    pub unsafe fn pool(&self) -> vk::CommandPool {
        self.pool.data_ptr().read()
    }
}

pub struct SingleUseCommand {
    device: Device,
    pool: Arc<SingleUseCommandPool>,
    buffer: vk::CommandBuffer,
}

impl !Sync for SingleUseCommand {}

impl Drop for SingleUseCommand {
    fn drop(&mut self) {
        unsafe {
            self.pool.pool.with_locked(|&pool| {
                self.device.dev().free_command_buffers(pool, &[self.buffer])
            })
        };
    }
}

impl SingleUseCommand {
    pub fn new(
        device: Device,
        pool: Arc<SingleUseCommandPool>,
    ) -> VkResult<Self> {
        let buffer = unsafe {
            let alloc_info = vk::CommandBufferAllocateInfo::default()
                .command_buffer_count(1)
                .command_pool(pool.pool())
                .level(vk::CommandBufferLevel::PRIMARY);
            let buffer = device.dev().allocate_command_buffers(&alloc_info)?[0];

            device.dev().begin_command_buffer(
                buffer,
                &vk::CommandBufferBeginInfo::default()
                    .flags(vk::CommandBufferUsageFlags::ONE_TIME_SUBMIT),
            )?;

            buffer
        };
        Ok(Self {
            device,
            pool,
            buffer,
        })
    }

    /// Safety: commandbuffer must not be accessed from multiple threads at the same time
    pub unsafe fn buffer(&self) -> vk::CommandBuffer {
        self.buffer
    }

    //pub fn copy_buffer_to_image(&self, image: &Image2D, buffer: &Buffer, )

    pub fn image_barrier(
        &self,
        image: vk::Image,
        aspects: vk::ImageAspectFlags,
        src_stage: vk::PipelineStageFlags2,
        src_access: vk::AccessFlags2,
        dst_stage: vk::PipelineStageFlags2,
        dst_access: vk::AccessFlags2,
        old_layout: vk::ImageLayout,
        new_layout: vk::ImageLayout,
        queue_ownership_op: Option<QueueOwnership>,
    ) {
        let (src_family, dst_family) = queue_ownership_op
            .map(|t| (t.src, t.dst))
            .unwrap_or((vk::QUEUE_FAMILY_IGNORED, vk::QUEUE_FAMILY_IGNORED));

        let barrier = vk::ImageMemoryBarrier2::default()
            .image(image)
            .subresource_range(
                vk::ImageSubresourceRange::default()
                    .aspect_mask(aspects)
                    .base_mip_level(0)
                    .base_array_layer(0)
                    .level_count(vk::REMAINING_MIP_LEVELS)
                    .layer_count(vk::REMAINING_ARRAY_LAYERS),
            )
            .src_stage_mask(src_stage)
            .src_access_mask(src_access)
            .dst_stage_mask(dst_stage)
            .dst_access_mask(dst_access)
            .dst_queue_family_index(dst_family)
            .src_queue_family_index(src_family)
            .old_layout(old_layout)
            .new_layout(new_layout);

        unsafe {
            self.device.dev().cmd_pipeline_barrier2(
                self.buffer,
                &vk::DependencyInfo::default()
                    .dependency_flags(vk::DependencyFlags::BY_REGION)
                    .image_memory_barriers(core::slice::from_ref(&barrier)),
            );
        }
    }

    pub fn clear_color_image(
        &self,
        image: vk::Image,
        format: vk::Format,
        layout: vk::ImageLayout,
        color: crate::Rgba,
        subresources: &[vk::ImageSubresourceRange],
    ) {
        let clear_colors = match format.get_component_kind() {
            crate::util::FormatComponentKind::Float => vk::ClearColorValue {
                float32: color.into_f32(),
            },
            crate::util::FormatComponentKind::UInt => vk::ClearColorValue {
                uint32: color.into_u32(),
            },
            crate::util::FormatComponentKind::SInt => vk::ClearColorValue {
                int32: color.into_i32(),
            },
        };

        unsafe {
            self.device.dev().cmd_clear_color_image(
                self.buffer,
                image,
                layout,
                &clear_colors,
                subresources,
            );
        }
    }

    pub fn submit_fence(
        &self,
        wait: Option<(vk::Semaphore, vk::PipelineStageFlags)>,
        signal: Option<vk::Semaphore>,
        fence: Option<vk::Fence>,
    ) -> VkResult<()> {
        unsafe { self.device.dev().end_command_buffer(self.buffer)? };

        let buffers = [self.buffer];
        let mut submit_info =
            vk::SubmitInfo::default().command_buffers(&buffers);

        if let Some(semaphore) = signal.as_ref() {
            // SAFETY: T and [T;1] have the same layout
            submit_info = submit_info.signal_semaphores(unsafe {
                core::mem::transmute::<&vk::Semaphore, &[vk::Semaphore; 1]>(
                    semaphore,
                )
            });
        }
        if let Some((semaphore, stage)) = wait.as_ref() {
            submit_info = submit_info
                .wait_semaphores(core::slice::from_ref(semaphore))
                .wait_dst_stage_mask(core::slice::from_ref(stage));
        }

        let fence = fence.unwrap_or(vk::Fence::null());
        self.pool.queue().with_locked(|queue| unsafe {
            self.device.dev().queue_submit(queue, &[submit_info], fence)
        })?;
        tracing::info!(
            "submitted queue {:?} and fence {:?}",
            self.pool.queue(),
            fence
        );

        Ok(())
    }

    pub fn submit_async<'a>(
        &'a self,
        wait: Option<(vk::Semaphore, vk::PipelineStageFlags)>,
        signal: Option<vk::Semaphore>,
        fence: Arc<sync::Fence>,
    ) -> VkResult<FenceFuture<'a>> {
        let device = self.device.clone();
        self.submit_fence(wait, signal, Some(fence.fence()))?;

        Ok(unsafe { FenceFuture::new(fence) })
    }

    pub fn submit_blocking(
        self,
        wait: Option<(vk::Semaphore, vk::PipelineStageFlags)>,
        signal: Option<vk::Semaphore>,
    ) -> VkResult<()> {
        let fence = Arc::new(sync::Fence::create(self.device.clone())?);
        let future = self.submit_async(wait, signal, fence)?;
        future.block();
        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    async fn async_submit(cmd: SingleUseCommand, queue: Queue) {
        cmd.submit_async(
            None,
            None,
            Arc::new(sync::Fence::create(cmd.device.clone()).unwrap()),
        )
        .unwrap()
        .await;
    }
}
